@@ -506,42 +506,60 @@ sub main {
 		}
 	}
 
-	# Select enhancing tasks for install.
-	my %provided;
+	# If an enhancing task is already marked for
+	# install, probably by preseeding, mark the tasks
+	# it enhances for install.
 	foreach my $task (grep { $_->{_install} && exists $_->{enhances} &&
 	                         length $_->{enhances} } @tasks) {
-		# If an enhancing task is already marked for
-		# install, probably by preseeding, mark the tasks
-		# it enhances for install.
 		map { $_->{_install}=1 } list_to_tasks($task->{enhances}, @tasks);
-		if (exists $task->{provides} && length $task->{provides}) {
-			$provided{$task->{provides}}=1;
-		}
 	}
-	foreach my $task (grep { ! $_->{_install} && exists $_->{enhances} &&
-	                         length $_->{enhances} } @tasks) {
-		my @deps=list_to_tasks($task->{enhances}, @tasks);
-		if (@deps) {
-			# Mark enhancing tasks for install if their
-			# dependencies are met and if their test fields
-			# mark them for install.
-			task_test($task, $options{"new-install"}, 0, 1);
-			foreach my $dep (@deps) {
-				if (! $dep->{_install}) {
-					$task->{_install} = 0;
+
+	# Select enhancing tasks for install.
+	# XXX FIXME ugly hack -- loop until enhances settle to handle
+	# chained enhances. This is ugly and could loop forever if
+	# there's a cycle.
+	my $enhances_needswork=1;
+	my %tested;
+	while ($enhances_needswork) {
+		$enhances_needswork=0;
+		foreach my $task (grep { ! $_->{_install} && exists $_->{enhances} &&
+		                         length $_->{enhances} } @tasks) {
+			my %tasknames = map { $_->{task} => $_ } @tasks;
+			my @deps=map { $tasknames{$_} } split ", ", $task->{enhances};
+
+			if (grep { ! defined $_ } @deps) {
+				# task enhances an unavailable or
+				# uninstallable task
+				next;
+			}
+
+			if (@deps) {
+				my $orig_state=$task->{_install};
+
+				# Mark enhancing tasks for install if their
+				# dependencies are met and their test fields
+				# mark them for install.
+				if (! exists $tested{$task->{task}}) {
+					$ENV{TESTING_ENHANCER}=1;
+					task_test($task, $options{"new-install"}, 0, 1);
+					delete $ENV{TESTING_ENHANCER};
+					$tested{$task->{task}}=$task->{_install};
+				}
+				else {
+					$task->{_install}=$tested{$task->{task}};
+				}
+
+				foreach my $dep (@deps) {
+					if (! $dep->{_install}) {
+						$task->{_install} = 0;
+					}
+				}
+
+				if ($task->{_install} != $orig_state) {
+					$enhances_needswork=1;
 				}
 			}
 		}
-
-		# If two enhancing tasks that both provide
-		# the same thing, only install one of them.
-		if ($task->{_install} && exists $task->{provides} &&
-		    length $task->{provides}) {
-			if (exists $provided{$task->{provides}}) {
-				$task->{_install}=0;
-			}
-			$provided{$task->{provides}}=1;
-		}
 	}
 
 	# Add tasks to install and see if any selected task requires manual
@@ -570,31 +588,21 @@ sub main {
 	else {
 		@aptitude="aptitude";
 	}
-
-	# Task removal..
-	if (@tasks_remove) {
-		my @packages_remove=map { task_packages($_, 0) } @tasks_remove;
-		foreach my $task (@tasks_remove) {
-			task_script($task->{task}, "prerm");
-		}
-		my $ret=run(@aptitude, "-y", "remove", @packages_remove);
-		if ($ret != 0) {
-			error gettext("aptitude failed")." ($ret)";
-		}
-		foreach my $task (@tasks_remove) {
-			task_script($task->{task}, "postrm");
-		}
-	}
 	
 	# And finally, act on selected tasks.
-	if (@tasks_install || $manual_selection) {
-		my @packages_install=map {task_packages($_, 1) } @tasks_install;
+	if (@tasks_install || @tasks_remove || $manual_selection) {
+		my @args;
+		foreach my $task (@tasks_remove) {
+			push @args, map { "$_-" } task_packages($task, 0);
+			task_script($task->{task}, "prerm");
+		}
 		foreach my $task (@tasks_install) {
+			push @args, task_packages($task, 1);
 			task_script($task->{task}, "preinst");
 		}
 		# If the user selected no other tasks and manual package
 		# selection, run aptitude w/o the --visual-preview parameter.
-		if (! @packages_install && $manual_selection) {
+		if (! @args && $manual_selection) {
 			my $ret=run("aptitude");
 			if ($ret != 0) {
 				error gettext("aptitude failed")." ($ret)";
@@ -602,15 +610,16 @@ sub main {
 		}
 		else {
 			if ($manual_selection) {
-				unshift @packages_install, "--visual-preview";
+				unshift @args, "--visual-preview";
 			}
-			my $ret=run(@aptitude, "--without-recommends",
-			                       "-y", "install",
-					       @packages_install);
+			my $ret=run(@aptitude, "-y", "install", @args);
 			if ($ret != 0) {
 				error gettext("aptitude failed")." ($ret)";
 			}
 		}
+		foreach my $task (@tasks_remove) {
+			task_script($task->{task}, "postrm");
+		}
 		foreach my $task (@tasks_install) {
 			task_script($task->{task}, "postinst");
 		}
